/*  ---------------------------------------------------------
	MPI performance test suite
	Transfer/2 - Test for MPI transfer rates/latency
	Laboratory of Parallel Information Technologies, SRCC MSU
	(parallel.ru/testmpi)
	Center for testing of supercomputer systems 
	(www.benchmark.ru)
	Copyright (c) 1997-2003

    this file: test_shell.c (generic test shell)
    to build a specific test, please attach
    a module that implements transfer_test_operation, 
    e.g.: mpicc transf1.c test_shell.c

    Last revision:  March, 31 / 2001
----------------------------------------------------------- */
   
#include "transfer.h"

static const char *copyright_message = "--- MPI Performance Test Suite\n--- Moscow State University 1998-2003";
extern const char test_description[];

static void* mpi_buf;
static const char *default_output_file_name = "transfer.dat";
static char output_file_name[256];
static int msglen_min = MIN_MESSAGE_LENGTH;
static int msglen_max = MAX_MESSAGE_LENGTH;
static int msglen_step = MESSAGE_INCREMENT;
static int msglen_multiplier = 1;
static int msglen_unit = 1;
static char unit_char = 'b';
static int buf_size = 0;
static int TIMES = DEFAULT_TIMES;
static int repeat = 1;
static int proc = MASTER; /* test transfer with all processes */
static int rank,pool_size,i_am_the_master;
static char print_mode = PRINT_RATE;
static int quiet = 0;
static int testcorrectness = 1;
static int fix_times = 0;
static double epsilon = 0.0;
int headproc = MASTER;

#define COMM MPI_COMM_WORLD


int main(int argc,char* argv[]) 
{
	int i,msglen;
	int cproc,ctimes;
	int repeatcnt;	
	int retest;
	double start_time,end_time;
	double t1,t2,rate,latency;
	void *sendbuf = NULL,*recvbuf = NULL;
	MPI_Status status;
	FILE *out = stdout;
	char processor_name[128];
	int namelen = 128;

	MPI_Init(&argc,&argv);
	MPI_Comm_size(COMM,&pool_size);
	MPI_Comm_rank(COMM,&rank);

	i_am_the_master = (rank == MASTER);	

	start_time = MPI_Wtime();

	if(i_am_the_master && !quiet)
	{
		fprintf(out,"%s\n\n",copyright_message);
		printf("\nRunning MPI Transfer/2 test, %d processes\n",pool_size);
	}

/* parse command line */
	if(i_am_the_master)	
	for(i = 1; i < argc; i ++)
		switch(argv[i][0]) 
		{
			case 'm': msglen_min = atoi(argv[i]+1); break;
			case 'M': msglen_max = atoi(argv[i]+1); break;
			case 's': msglen_step = atoi(argv[i]+1); msglen_multiplier = 1; break;
			case 'K': msglen_multiplier = atoi(argv[i]+1); msglen_step = 0; break;
			case 'b': buf_size = atoi(argv[i]+1); break;
			case 'T': TIMES = atoi(argv[i]+1); break;			
			case 'R': repeat = atoi(argv[i]+1); break;			
			case 'f': fix_times = atoi(argv[i]+1); epsilon = (double) fix_times / 1000.0; break;			
			case 'p': proc = atoi(argv[i]+1); break;			
			case 'H': headproc = atoi(argv[i]+1); break;			
			case 'P': print_mode = argv[i][1]; if(!quiet) printf("Output mode set to %s\n",argv[i]+1); break;			
			case 'u': switch(argv[i][1])
					  {
						case 'M':
						case 'm': msglen_unit = MEGABYTE; unit_char='M'; break;
						case 'K':
						case 'k': msglen_unit = KILOBYTE; unit_char='K'; break;
						case 'B':
						case 'b': msglen_unit = 1; unit_char='b'; break;
					  }
						break;

			case 'q': quiet = 1; break;			
			case 'o': 
					strcpy(output_file_name,argv[i]+1);  
					out = fopen(output_file_name,"a+");
					if(out == NULL)
					{
						printf("WARNING: Can't open or create %s, writing to stdout\n",output_file_name);
						out = stdout;
					}
					else 
						printf("Writing output to %s\n",output_file_name);
               		break;

			default: fprintf(stderr,"WARNING: unrecognized option: %s\n",argv[i]);
			break;
		}

	/* exchange parameters */

		msglen_min  *= msglen_unit;
		msglen_max  *= msglen_unit;
		msglen_step *= msglen_unit;

	if(i_am_the_master && !quiet) 
		fprintf(out,"messages: %d to %d, step %d, unit is %d bytes;\n %d times (fix %d)\n",
         msglen_min,msglen_max,msglen_step,msglen_unit,TIMES,fix_times,buf_size);		

		MPI_Bcast(&msglen_min,1,MPI_INT,MASTER,COMM);
		MPI_Bcast(&msglen_max,1,MPI_INT,MASTER,COMM);
		MPI_Bcast(&msglen_step,1,MPI_INT,MASTER,COMM);
		MPI_Bcast(&msglen_multiplier,1,MPI_INT,MASTER,COMM);
		MPI_Bcast(&msglen_unit,1,MPI_INT,MASTER,COMM);
		MPI_Bcast(&buf_size,1,MPI_INT,MASTER,COMM);		
		MPI_Bcast(&TIMES,1,MPI_INT,MASTER,COMM);		
		MPI_Bcast(&repeat,1,MPI_INT,MASTER,COMM);		
		MPI_Bcast(&proc,1,MPI_INT,MASTER,COMM);		
		MPI_Bcast(&fix_times,1,MPI_INT,MASTER,COMM);		
		MPI_Bcast(&epsilon,1,MPI_DOUBLE,MASTER,COMM);		
		MPI_Bcast(&headproc,1,MPI_DOUBLE,MASTER,COMM);		

    	i_am_the_master = (rank == headproc) ? 1 : 0;
			
	
	if(!quiet)
	{
		MPI_Barrier(MPI_COMM_WORLD);
        MPI_Get_processor_name(processor_name,&namelen);
		if(i_am_the_master)
		{
		    fprintf(out,"Process %d of %d on %s\n",rank, pool_size, processor_name);
			for(i = 0; i < pool_size; i ++)
			if(i != headproc)
			{
			     MPI_Recv(processor_name,128,MPI_CHAR,i,17,MPI_COMM_WORLD,&status);
                             fprintf(out,"Process %d of %d on %s\n",i, pool_size, processor_name);
			}
		}
		else
			MPI_Send(processor_name,128,MPI_CHAR,headproc,17,MPI_COMM_WORLD);

		fflush(out);
	}


	if(buf_size != 0) 
	{
		mpi_buf = malloc(buf_size);
		if(mpi_buf == NULL)
		{ 
			fprintf(stderr,"\nFailed to allocate MPI buffer of size %d\n"); 
			MPI_Abort(COMM,1);
			exit(1); 
		}

		MPI_Buffer_attach(mpi_buf,buf_size);
	}
	

	if(i_am_the_master && !quiet) 
	{

		fprintf(out,"\n\nRunning test: %s\n",test_description);
		if(proc != headproc)
			fprintf(out,"Testing transfer between %d and %d\n\n",headproc,proc);
		else
			fprintf(out,"Testing transfer between %d and other processes\n\n",headproc);
		if(testcorrectness)
			fprintf(out,"Checking correctness of message passing\n");			
		fflush(out);
	 }

	if(i_am_the_master)
	{
		if(print_mode == 'r')
			fprintf(out,"\nSize(%c)\tTransfer (MB/sec)\n\n",unit_char);
		else
			fprintf(out,"\nSize(%c)\tTime (microsec)\n\n",unit_char);
		fflush(out);

	}
	for(repeatcnt = 0; repeatcnt < repeat; repeatcnt ++)
	{
	if(i_am_the_master) fprintf(out,"\nIteration %d\n",repeatcnt);

	for(msglen = msglen_min; msglen <= msglen_max; msglen = msglen * msglen_multiplier + msglen_step) 
	{
		if(msglen) 
		{
        	sendbuf = malloc(msglen);
        	recvbuf = malloc(msglen);
            if(sendbuf == NULL || recvbuf == NULL) 
			{
				fprintf(stderr,"Can't allocate %d bytes\n",msglen);
				exit(1);
            }
		}	


		if(i_am_the_master) 
		{
			if((msglen != 0) || (print_mode == PRINT_TIME)) 
			{
				if(msglen_unit > 1)
		 			fprintf(out,"%d%c\t",msglen/msglen_unit,unit_char);
				else
		 			fprintf(out,"%d\t",msglen/msglen_unit);
			}
		}

		/* Start loop over processes */

		for(cproc = 0; cproc < pool_size; cproc ++)
		{
			if(cproc == headproc)
				continue;

			if(testcorrectness )
			{
				fillbuffer(sendbuf,msglen,'A');
				fillbuffer(recvbuf,msglen,'B');

				transfer_test_operation(sendbuf,recvbuf,msglen,rank,1,cproc);				

				if(rank == headproc || rank == cproc)
				{
					if(!checkbuffer(recvbuf,msglen,'A'))
					{
						printf("[%d] receive buffer of size %d is not correct after operation\n",rank,msglen);
						MPI_Abort(COMM,1);
						exit(1);
					}
				}
				else
				{
					if(!checkbuffer(recvbuf,msglen,'B'))
					{
						printf("[%d] receive buffer of size %d is not correct after operation (changed!)\n",rank,msglen);
						MPI_Abort(COMM,1);
						exit(1);
					}
				}
			}

			/* End of testing correctness */

			ctimes = TIMES;
			retest = 0;

			do
			{
	            MPI_Barrier(COMM);
				t2 = t1 = MPI_Wtime();
				transfer_test_operation(sendbuf,recvbuf,msglen,rank,ctimes,cproc);
				MPI_Barrier(COMM);
				t2 = MPI_Wtime();
        
				if(i_am_the_master && fix_times)
				{
				if(t2 - t1 < epsilon)
				{
					ctimes = ctimes * 2;
					retest = 1;
				}
				else
				{
					retest = 0;
				}
				}
				MPI_Bcast(&retest,1,MPI_INT,headproc,COMM);
				MPI_Bcast(&ctimes,1,MPI_INT,headproc,COMM);
			}	while(retest);

			if(i_am_the_master) 
			{
				
				if(msglen != 0 || print_mode == PRINT_TIME) 
				{
					if(print_mode == PRINT_RATE)
					{
						if(t2 == t1)
						{
							fprintf(out,"T=0!\t");
						}	
						else
						{
							rate = ((double)msglen * 2.0 * (double) ctimes)/((t2 - t1) * (double) MEGABYTE);
					 		fprintf(out,"%03.4g\t",rate);
						}
					}
					else
						fprintf(out,"%03.4g\t",1000000.0*(t2-t1)/(2.0*ctimes));
				}
				else 
				{
					latency = (t2 - t1) * 1000000 / ((double) ctimes * 2);
					fprintf(out,"[%d -- %d] Latency: %g microseconds (at %d times)\n",headproc,cproc,latency,ctimes);
				} 
    
			fflush(stdout);
			fflush(out);
			} /* end of master's section */
		}

		if(i_am_the_master)
			fprintf(out,"\n");

		if(msglen) 
		{
			free(sendbuf);
			free(recvbuf);
		}

	}  /* end of loop over message sizes */
	} /* end of repeat loop */
	
	if(out != stdout) 
		fclose(out);

	end_time = MPI_Wtime();

	if(i_am_the_master && !quiet)
		printf("\nMPI Transfer test complete in %g seconds\n",end_time-start_time);

	MPI_Finalize();

	return 0;	
}

void fillbuffer(char *buffer,int size,char c)
{
 	int i;
	for(i = 0; i < size; i ++)
		buffer[i] = c;
}

int checkbuffer(char *buffer,int size,char c)
{
 	int i;
	for(i = 0; i < size; i ++)
		if(buffer[i] != c)
			return 0;
	return 1;
}
